/*
 * din irc bot
 * based on LGPLed code by Georgy Yunaev gyunaev@ulduzsoft.com 
 * of libircclient.sourceforge.net.
 *
*/

#include <cstdio>

#include "bot.h"
#include "console.h"
#include "command.h"
#include "tokenizer.h"

#include <iostream>
using namespace std;

extern console cons;
extern cmdlist cmdlst;

bot::bot (const string& ln, const string& sn, const string& hlp) : command (ln, sn, hlp), sess (0), tid(0), prefixed (0) {

		memset (&callbacks, 0, sizeof(callbacks));

		callbacks.event_connect = event_connect;
		callbacks.event_join = event_join;
		callbacks.event_nick = dump_event;
		callbacks.event_quit = dump_event;
		callbacks.event_part = dump_event;
		callbacks.event_mode = dump_event;
		callbacks.event_topic = dump_event;
		callbacks.event_kick = dump_event;
		callbacks.event_channel = event_channel;
		callbacks.event_privmsg = event_privmsg;
		callbacks.event_notice = dump_event;
		callbacks.event_invite = dump_event;
		callbacks.event_umode = dump_event;
		callbacks.event_ctcp_rep = dump_event;
		callbacks.event_ctcp_action = dump_event;
		callbacks.event_unknown = dump_event;
		callbacks.event_numeric = event_numeric;
		callbacks.event_dcc_chat_req = irc_event_dcc_chat;
		callbacks.event_dcc_send_req = irc_event_dcc_send;

		sess = irc_create_session (&callbacks);
		
		if (!sess) {
			cout << "could not create IRC session" << eol;
		} else {
			cout << "created an IRC session" << eol;
			irc_set_ctx (sess, this);
		}

}
	
bool bot::operator() (tokenizer& tz) {
	
	string cmd;
	tz >> cmd;
	if ((cmd == "prefix") || (cmd == "p")) {
			string value; tz >> value; 
			if (value != "") {
				prefix = value + ':';
				prefixed = 1;
				cons << console::green << "will prefix: " << prefix << eol; 
			} else { 
				cons << console::red << "wont prefix: " << prefix << eol; 
				prefix = "";
				prefixed = 0; 
			}
	} else if ( (cmd == "connect") || (cmd == "c")) {
		tz >> srv >> nick >> chan;
		// add # to channel name
		string s = "#"; s += chan; chan = s;
		nickolon = nick + ':';
		len = nickolon.length ();
		kill ();
		make ();
	} else { // send as message to channel
		string msg(cmd), wrd;
		while (1) {
			tz >> wrd;
			if (wrd.length()) msg = msg + ' ' + wrd; else break;
		} 
		if (msg.length()) *this << msg;
	
	}
	return true;
}

void* irc_monitor (void* pb) {
	bot* b = (bot*) pb;
	irc_run (b->sess);
	return 0;
}

int bot::make () {
	
	if (sess && !tid) { // no previous bot
	
		if ( int e = irc_connect (sess, srv.c_str(), 6667, 0, nick.c_str(), 0, 0) ) {
			cons << console::red << srv << " not found. " << irc_strerror(e) << eol;
			return 0;
		} else cons << console::cyan << "connecting to " << srv << eol;
				
		pthread_create (&tid, 0, irc_monitor, this);
	
		return 1;
		
	}
	
	return 0;
	
}

int bot::kill () {
	if (tid) {
		irc_cmd_quit (sess, "i'm done.");
		pthread_join (tid, 0);
		tid = 0;
		if (sess) irc_disconnect (sess);
		return 1;
	} 
	return 0;
}

bot& bot::operator<< (string& s) {
	if (tid && (s != "")) {
		string out (prefix + s.c_str());
		irc_cmd_msg (sess, chan.c_str(), out.c_str());
	}
	return *this;
}

bot::~bot () {
	kill ();
	if (sess) irc_destroy_session (sess);
}

void dump_event (irc_session_t * session, const char * event, const char * origin, const char ** params, unsigned int count){}

void event_join (irc_session_t * session, const char * event, const char * origin, const char ** params, unsigned int count)
{
	irc_cmd_user_mode (session, "+b");
	cons << console::yellow << "joined " << params[0] << " as " <<  get_nick (origin) << eol;
}

void event_connect (irc_session_t * session, const char * event, const char * origin, const char ** params, unsigned int count)
{
	bot* pbot = (bot *) irc_get_ctx (session);
	cons << console::green << "connected." << eol;
	irc_cmd_join (session, pbot->chan.c_str(), 0);
}

string get_nick (const char * origin) {
	
	static const int nickbufsize = 512;
	static char nickbuf [nickbufsize]; // irc_get_target_nick will truncate to this length
	irc_target_get_nick (origin, nickbuf, nickbufsize);
	return string (nickbuf);
	
}

void event_privmsg (irc_session_t * session, const char * event, const char * origin, const char ** params, unsigned int count)
{
	bot* pb = (bot *) irc_get_ctx (session);
	cmdlst.run (params[1]);
	*pb << cmdlst.result;
}

void event_channel (irc_session_t * session, const char * event, const char * origin, const char ** params, unsigned int count)
{
	bot* pb = (bot *) irc_get_ctx (session);
	pb->mesg = params[1];
	string tok (pb->mesg.substr (0, pb->len));
	if (tok == pb->nickolon) {
		pb->mesg = pb->mesg.substr (pb->len, pb->mesg.length());
		cmdlst.run (pb->mesg);
		*pb << cmdlst.result;
	} else {
		cons << get_nick (origin) << "@" << pb->srv << ": " << pb->mesg << eol;
	}
}

void irc_event_dcc_chat (irc_session_t * session, const char * nick, const char * addr, irc_dcc_t dccid){}
void irc_event_dcc_send (irc_session_t * session, const char * nick, const char * addr, const char * filename, unsigned long size, irc_dcc_t dccid){}
void event_numeric (irc_session_t * session, unsigned int event, const char * origin, const char ** params, unsigned int count) {}
