#include "globals.h"
#include "din.h"
#include "command.h"
#include "console.h"
#include "math.h"
#include "delay.h"
#include "scalelist.h"
#include "font.h"
#include "glyph.h"
#include "chrono.h"
#include "utils.h"
#include "ui.h"
#include "curve_editor.h"
#include "bot.h"
#include "tokenizer.h"
#include "tcl_interp.h"
#include "chrono.h"

#define buf cmdlst.buf
#define result cmdlst.result

using namespace std;

extern console cons;
extern bot bott;
extern cmdlist cmdlst;

void cmdlist::add (command* cmd) {
	cmds.push_back (cmd);
	ncmds = cmds.size ();
}

command* cmdlist::run (const string& cmd) {
	
	tokenizer tz (cmd);		
	string name; tz >> name;
	
	result = "";
	for (int i = 0; i < ncmds; ++i) {
		command* ci = cmds[i];
		if ((name == ci->longname) || (name == ci->shortname)) {
			(*ci)(tz);
			return ci;
		}
	}
	
	return 0;
	
}

bool cmdlist::f1 (const string& cmd) {
	
	// help for one command
	for (int i = 0, j = cmds.size(); i < j; ++i) {
		command* ci = cmds[i];
		if ((cmd == ci->longname) || (cmd == ci->shortname)) {
			ci->f1();
			return true;
		}
	}
	result = cmd + " not found";
	return false;
}

void cmdlist::list () {
	// list all available commands
	cons << console::green << "Commands available:" << eol;
	string comma = ", ";
	for (int i = 0, cmds_ln = 5, j = cmds_ln; i < ncmds; ++i) {
		cons << console::cyan << cmds[i]->longname << comma;
		if (j--); else {
			j = cmds_ln;
			cons << eol;
		}  
	}
	cons << console::green << "Type help <command_name> to find out more." << eol;
	
}

bool setv::operator() (tokenizer& tz) {
	string varn; tz >> varn;
	if (varn == "sc" || varn == "scroll") {
		string x, y; tz >> x >> y;
		int ix = d->dinfo.scroll.x, iy = d->dinfo.scroll.y;
		if (x != ".") ix = atoi (x.c_str()); 
		if (y != ".") iy = atoi (y.c_str());
		d->dinfo.scroll = point<int>(ix, iy);
	} else if (varn == "fmd" || varn == "fm_depth") {
		tz >> d->fm_depth;
		hz2step (d->fm_depth, d->fm_step);
	} else if (varn == "amd" || varn == "am_depth") {
		tz >> d->am_depth;
	} else if (varn == "snap_drones" || varn == "sd") {
		tz >> d->snap_drones;
	} else if (varn == "tu" || varn == "tuning") {
		extern load_globals lg;
		string name; tz >> name;
		if (lg.load_intervals (name)) {
			result = "set tuning to: " + name;
		} else {
			result = "bad tuning file: " + name;
			return false;
		}
		d->restore_all_notes ();
	} else if (varn == "ps1" || varn == "prompt") {
		tz >> cons.ps1;
		cons.calc_draw_params ();
		return true;
	}	else {
		result = "variable " + varn + " not found ";
		return false;
	}
	
	return true;
}

bool getv::operator() (tokenizer& tz) {
	string varn; tz >> varn;
	static const char* ff = "%.3f";
	if (varn == "key" || varn == "tonic") {
		sprintf (buf, ff, d->get_tonic ());
		result = buf;
		return true; 
	} else if (varn == "sc" || varn == "scroll") {
		sprintf (buf, "%d %d", d->dinfo.scroll.x, d->dinfo.scroll.y);
		result = buf;
		return true;
	} else if (varn == "fmd" || varn == "fm_depth") {
		sprintf (buf, ff, d->fm_depth);
		result = buf;
		return true;
	} else if (varn == "amd" || varn == "am_depth") {
		sprintf (buf, ff, d->am_depth);
		result = buf;
		return true;
	} else if (varn == "sd" || varn == "snap_drones") {
		sprintf (buf, "%d", d->snap_drones);
		return true;
	} else {
		return false;
	}
}

bool show_help::operator() (tokenizer& tz) {
	string cmd; tz >> cmd;
	if (cmd.length()) lst->f1(cmd); else lst->list ();
	return true;
}

bool set_delay::operator() (tokenizer& tz) {
	
	string name;
	float value;

	delay* d[2] = {left, right};
	delay** pd = d;
	int n = -1;
	
	tz >> name >> value;
	if (name == "a" || name == "all") n = 2; 
	else if (name == "l" || name == "left") n = 1; 
	else if (name == "r" || name == "right") {
		pd++; 
		n = 1;
	} else {
		result = "bad delay name";
		return false;
	}
		
	for (int i = 0; i < n; ++i, pd++) {
		(*pd)->set (value);
	}

	return true;
	
}

bool get_delay::operator() (tokenizer& tz) {
	
	delay* d;
	string s; tz >> s;
	if (s == "left" || s == "l") d = left; 
	else if (s == "right" || s == "r") d = right; 
	else return false;
	
	float t;
	d->get (t);
	
	sprintf (buf, "%0.2f", t);
	result = buf;
	
	return true;
	
}

extern din din0; 
const char* bpm_com::str [bpm_com::NUM] = {"gl", "gr", "am", "fm", "os"};
beat2value* bpm_com::bv [] = { &din0.gatl, &din0.gatr, &din0.am, &din0.fm, &din0.octave_shift };

void bpm_com::fill (const string& name) {
	
	setset(0);
	if (name == "g" || name == "gate") set[0]=set[1]=1; else if (name == "all" || name == "a") setset(1); else {
		for (int i = 0; i < NUM; ++i) {
			if (name == str[i]) {
				set[i] = 1;
				break;
			}
		}
	}	
	
}

bool set_bpm::operator() (tokenizer& tz) {
	
	string name;
	float val;
	tz >> name >> val;
	fill (name);
	for (int i = 0; i < NUM; ++i) if (set[i]) {
		bv[i]->set_bpm (val);
	}

	return true;
	
}

bool get_bpm::operator() (tokenizer& tz) {
	
	string name; tz >> name;
	fill (name);
	
	for (int i = 0; i < NUM; ++i) {
		if (set[i]) {
			sprintf (buf, "%.3f", bv[i]->get_bpm());
			result = buf;
		}
	}
	return true;
	
}

bool set_style::operator() (tokenizer& tz) {
	
	string name, style; 
	tz >> name >> style;
	
	fill (name);
	
	xhandler *xmin = 0, *xmax = 0;
	if (style == "loop" || style == "l") {
		xmin = &_loopmin;
		xmax = &_loopmax;
	} else if (style == "pong" || style == "p") {
		xmin = &_pongmin;
		xmax = &_pongmax;
	} else return false;
	
	for (int i = 0; i < NUM; ++i) {
		if (set[i]) {
			beat2value* bvi = bv[i];
			bvi->xmin = xmin;
			bvi->xmax = xmax;
			bvi->style = style;
		}
	}
	
	return true;
	
}

bool get_style::operator() (tokenizer& tz) {
	string name; tz >> name;
	for (int i = 0; i < NUM; ++i) {
		if (name == str[i]) result = bv[i]->style;
	}
	return true;
}

bool gaters::operator() (tokenizer& tz) {
	string target, state; tz >> target >> state;
	if (state == "on") {
		set_state (target, true);
		result = state;
		return true;
	} else if (state == "off") {
		set_state (target, false);
		result = target + ' ' + state;
		return true;
	} else {
		result = get_state (target);
		return false;
	}
}

void gaters::set_state (const string& target, const bool& state) {
	if (target == "gl" || target == "gate-l") d->gatl.enabled = state;
	else if (target == "gr" || target == "gate-r") d->gatr.enabled = state;
}

string gaters::get_state (const string& target) {
	bool state = false;
	if (target == "gl" || target == "gate-l") state = d->gatl.enabled;
	else if (target == "gr" || target == "gate-r") state = d->gatr.enabled;
	if (state) return "on"; else return "off";
}

bool add_scale::operator() (tokenizer& tz) {
	extern scalelist scalelst;
	scalelst.add (tz);
	return true;
}

bool ls_scale::operator() (tokenizer& tz) {
	extern scalelist scalelst;
	scalelst.ls ();
	return true;
}

bool remove_scale::operator() (tokenizer& tz) {
	string nam; tz >> nam;
	if (nam != "") {
		extern scalelist scalelst;
		scalelst.remove (nam);
		return true;
	} 
	return false;
}

bool load_scale::operator() (tokenizer& tz) {
	string nam; tz >> nam;
	if (nam != "") {
		extern scalelist scalelst;
		scale r;
		if (scalelst.get (nam, r)) {
			d->save_scale ();
			d->scaleinfo.set_scale (nam);
			d->load_scale ();
			d->enter ();
			return true;
		} else {
			cons << console::red << "scale: " << nam << " not found." << eol;
			return false;
		}
	} else {
		cons << console::red << "bad scale name" << eol;
		return false;
	}
}

bool set_kern::operator() (tokenizer& tz) {
	
	extern font fnt;
	
	char a, b;
	int k;
	
	tz >> a >> b >> k;
	
	if (a == '.') {
		map<char, glyph>& m = fnt.characters;
		for (map<char, glyph>::iterator i = m.begin(), j = m.end(); i != j; ++i) fnt.kern[(*i).first][b] = k * fnt.xsize;
		fnt.modified = 1;
	} else if (b == '.') {
		map<char, glyph>& m = fnt.characters;
		for (map<char, glyph>::iterator i = m.begin(), j = m.end(); i != j; ++i) fnt.kern[a][(*i).first] = k * fnt.xsize;
		fnt.modified = 1;
	} else {
		fnt.kern[a][b] = k * fnt.xsize;
		fnt.modified = 1;
	}
	
	return true;
	
}

bool get_kern::operator() (tokenizer& tz) {
	
	extern font fnt;
	
	char a, b;
	int k;
	
	tz >> a >> b >> k;
	
	if (a == '.') {
		map<char, glyph>& m = fnt.characters;
		for (map<char, glyph>::iterator i = m.begin(), j = m.end(); i != j; ++i) {
			a = (*i).first;
			cons << a << ' ' << b << ' ' << fnt.kern[a][b] / fnt.xsize << eol;
		}
	} else if (b == '.') {
		map<char, glyph>& m = fnt.characters;
		for (map<char, glyph>::iterator i = m.begin(), j = m.end(); i != j; ++i) {
			b = (*i).first;
			cons << a << ' ' << b << ' ' << fnt.kern[a][b] / fnt.xsize << eol;
		}
	} else {
		cons << a << ' ' << b << ' ' << fnt.kern[a][b] / fnt.xsize << eol;
	}
	
	return true;
	
}

bool set_font_size::operator() (tokenizer& tz) {
		
	extern font fnt;
	
	int xs, ys;
	tz >> xs >> ys;
	
	if (xs < 1) xs = fnt.xsize;
	if (ys < 1) ys = fnt.ysize;
	fnt.cell_size (xs, ys);
	fnt.modified = 1;
	
	extern console cons;
	cons.calc_draw_params ();
		
	return true;
	
}

bool get_font_size::operator() (tokenizer& tz) {
	cons << "xsize " << fnt.xsize << " ysize " << fnt.ysize << eol;
	return true;
}

bool load_resonators ::operator() (tokenizer& tz) {
	string fname; tz >> fname;
	extern audio_out aout;
	aout.stop ();
		if (fname == ".") d->load_resonators (d->dinfo.resonators_fname); else d->load_resonators (fname);
	aout.start ();
	return true;
}

bool save_resonators ::operator() (tokenizer& tz) {
	string fname; tz >> fname;
	if (fname != ".") d->dinfo.resonators_fname = fname;
	int deldis = 0; tz >> deldis;
	d->save_resonators (deldis);
	return true;
}

bool delete_disabled_resonators ::operator() (tokenizer& tz) {
	extern cmdlist cmdlst;
	cmdlst.run ("sre . 1");
	cmdlst.run ("lre .");
	return true;
}

const char* western1 [] = {
	"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C"
};

const char* western2 [] = {
	"C", "Db", "D", "Eb", "E", "F", "F#", "G", "Ab", "A", "Bb", "B", "C"
};

const char* indian [] = {
	"s", "r", "R", "g", "G", "m", "M", "P", "d", "D", "n", "N", "S"
};

const int NUM_NOTES = 13;

const float frequencies [] = { // from wikipedia - piano key frequencies page.
	261.626, 277.183, 293.665, 311.127, 329.628, 349.228, 369.994, 391.995, 415.305, 440, 466.164, 493.883, 523.251
};

int get_note_num (const string& s) {
	for (int i = 0; i < NUM_NOTES; ++i) {
		if (s == western2[i] || s == western1[i]) return i;
	}
	cons << console::red << "bad note: " << s << eol;
	return -1;
}

bool note_distance::operator() (tokenizer& tz) {
	
	string note1, note2;
	tz >> note1 >> note2;
	if (note1.length() && note2.length()) {
		note1[0] = toupper (note1[0]);
		note2[0] = toupper (note2[0]);
	} else return false;
	
	static const char* dist [] = {
		"same note", "minor second", "second", "minor third", "third", "perfect fourth",
		"diminished fifth", "perfect fifth", "minor sixth", "sixth", "minor seventh",
		"seventh", "octave"
	};
	
	int i = get_note_num (note1), j = get_note_num (note2);
	if (i == -1 || j == -1) return false; else {
		if (j <= i) j += 12;
		int ji = j - i;
		cons << dist[ji] << eol;
	}
	
	return true;
	
}

bool chord::operator() (tokenizer& tz) {
	
	string root, type;
	tz >> root >> type;
	
	if (root.length()) root[0] = toupper (root[0]); else return false;
	int nn = get_note_num (root);
	if (nn == -1) return false; else {
		
		const int ntypes = 20;
		static const char* types[] = {
			"", // major
			"m", // minor
			"maj7", // major 7
			"m7", // minor 7
			"sus4", // suspended 4th
			"7sus4", // dominant 7th, suspended 4th
			"6", // major 6th
			"m6", // minor 6th
			"7", // dominant 7th
			"9", // dominant 9th
			"5", // power chord
			"69", // maj 6th, add 9
			"11", // dominant 11th
			"13", // dominant 13th
			"add9", // major add 9th
			"m9", // minor 9th
			"maj9", // major 9th
			"+", // augmented
			"07", // diminished 7th
			"0" // diminished triad
		};
		
		static const int steps[][7] = {
			{3, 0, 4, 7}, // num_notes, root, note1, note2....
			{3, 0, 3, 7},
			{4, 0, 4, 7, 11},
			{4, 0, 3, 7, 10},
			{3, 0, 5, 7},
			{4, 0, 5, 7, 10},
			{4, 0, 4, 7, 9},
			{4, 0, 3, 7, 9},
			{4, 0, 4, 7, 10},
			{5, 0, 4, 7, 10, 14},
			{2, 0, 7},
			{5, 0, 4, 7, 9, 14},
			{6, 0, 4, 7, 10, 14, 17},
			{6, 0, 4, 7, 10, 14, 21},
			{4, 0, 4, 7, 14},
			{5, 0, 3, 7, 10, 14},
			{5, 0, 4, 7, 11, 14},
			{3, 0, 4, 8},
			{4, 0, 3, 6, 10},
			{3, 0, 3, 6}
		};
		
		int t = -1;
		for (int i = 0; i < ntypes; ++i) {
			if (type == types[i]) {
				t = i; break;
			}
		}
		
		if (t == -1) return false; else {
			const int* ch = &steps[t][0];
			stringstream ss1, ss2;
			for (int i = 0, j = ch[0], k = 1; i < j; ++i, ++k) {
				int p = (nn + ch[k]) % 12;
				ss1 << western1[p] << ' ';
				ss2 << western2[p] << ' ';
			}
			cons << ss2.str() << " or " << ss1.str() << eol;
			return true;
		}
		
		
	}
}

void key::find_nearest_note (string& note, float& frequency, float& dist) {
	
	float left = frequencies[0] / 16, right = 2 * left;
	float tonic = d->get_tonic ();
	if (tonic < left) return;
	
	while (1) {
		if ((left <= tonic) && (right > tonic)) break;
		else { 
			left*=2; 
			right*=2;
		}
	}
	
	float oct = left / frequencies[0];
	dist = tonic - left;
	int id = 0; 
	float tone = 0;
	for (int i = 0; i < 13; ++i) {
		tone = frequencies[i] * oct;
		float newdist = tone - tonic; if (newdist < 0) newdist = -newdist;
		if (newdist < dist) {
			dist = newdist;
			id = i;
		}
	}	
	
	string nn;
	if (western2[id] == western1[id]) nn = western1[id]; else nn = (string)western2[id] + "/" + (string)western1[id];
	note = nn;
	frequency = frequencies[id] * oct;
	dist = tonic - frequency;
	d->set_key_id (id);

}

bool key::operator() (tokenizer& tz) {
	
	string s; tz >> s;
	if (s == "") {
		// no args, just print key
		string note; 
		float frequency, dist;
		find_nearest_note (note, frequency, dist);
		sprintf (buf, "tonic: %0.3f Hz, nearest_note: %s/%0.3f Hz, distance: %0.3f Hz", d->get_tonic(), note.c_str(), frequency, dist);
		result = buf;
		return true;
	} else {
		float f = d->get_tonic ();
		int k = d->get_key_id ();
		int o = 0;
		float v = atof (s.c_str());
		if (v <= 0) { // a note
			if (s[0] != '.') {
				s[0] = toupper(s[0]);
				for (int i = 0; i < 12; ++i) {
					if ((s == western1[i]) || (s == western2[i])) {
						f = frequencies[i];
						k = i;
						break;
					}
				}
			}
		} else { // a value
			f = v;
		}
		
		// octave
		tz >> s;
		o = atoi (s.c_str());
				
		// check and set
		f = f * pow (2, o);
		bool tonicok = check (f);
		if (tonicok) {
			d->set_tonic (f);
			d->set_key_id (k);
			return true;
		}
		
	}

	return false;

}

bool key::check (float f) {
	const float min_key = frequencies[0] / 8, max_key = 20000;
	if (f < min_key) {
		result = "tonic too low";
		cons << console::red << result << eol;
		return false;
	} else if (f > max_key) {
		result = "tonic too high";
		cons << console::red << result << eol;
		return false;
	} else return true;
}

bool notation::operator () (tokenizer& tz) {
	
	string s;
	tz >> s;
		
	if (s == "numeric" || s == "n") d->set_notation ("numeric"); else
	if (s == "w" || s == "west" || s == "western") d->set_notation ("western");
	
	return true;
	
}

bool echo::operator() (tokenizer& tz) {
	string s, ls;
	string spew;
	while (1) {
		tz >> s;
		if (s == ls) break;
		spew = spew + s + ' ';
		ls = s;
	} 
	cons << console::cyan << spew << eol;
	return true;
}

bool set_pitch_and_volume::operator() (tokenizer& tz) {
	
	float pitch, volume;
	string ignore;
	
	tz >> pitch >> volume;
	
	extern int mousex, mousey;
	extern viewport view;
	
	mousex = pitch * view.xmax;
	mousey = volume * view.ymax;
	
	cons << pitch << ' ' << volume << eol;
		
	return true;
	
}

bool mouse::operator() (tokenizer& tz) {
	
	int x, y; tz >> x >> y;
	
	string sx, sy;
	tz >> sx >> sy;
	
	extern int mousex, mousey;
	SDL_WarpMouse (mousex + x, mousey + y);
	
	return true;
	
}

extern uis_t uis;

bool curve_value::operator() (tokenizer& tz) {
	curve_editor* ced = uis.crved;
	if (ced && ced->pik()) {
		string sx, sy;
		tz >> sx >> sy;
		hit_t& h = ced->pik;
		multi_curve* crv = ced->curveinfo[h.crv_id].curve;
		float x, y;
		switch (h.what) {
			case hit_t::VERTEX:
				crv->get_vertex (h.id, x, y);
				parse (sx, sy, x, y);
				ced->move (h, x, y);
				return true;
			case hit_t::LEFT_TANGENT:
				crv->get_left_tangent (h.id, x, y);
				parse (sx, sy, x, y);
				ced->move (h, x, y);
				return true;
			case hit_t::RIGHT_TANGENT:
				crv->get_right_tangent (h.id, x, y);
				parse (sx, sy, x, y);
				ced->move (h, x, y);
		}
		return false;
	}
	return false;
}

void curve_value::parse (const string& sx, const string& sy, float& x, float& y) {
	if (sx != ".") x = atof (sx.c_str());
	if (sy != ".") y = atof (sy.c_str());
}

bool curve_name::operator() (tokenizer& tz) {
	curve_editor* ced = uis.crved;
	if (ced && ced->pik()) {
		multi_curve* crv = ced->curveinfo[ced->pik.crv_id].curve;
		tz >> crv->name;
	}
	return true;
}

bool curve__library::operator() (tokenizer& tz) {
	curve_editor* ced = uis.crved;
	if (!ced) {
		result = "not a curve editor";
		return false;
	}
	curve_library* cl = ced->library;
	if (cl) {
		string cmd; tz >> cmd;
		if (cmd == "a" || cmd == "add") ced->add_curve (); else
		if (cmd == "d" || cmd == "delete") cl->del (); else
		if (cmd == "u" || cmd == "update") ced->replace_curve (); else
		if (cmd == "i" || cmd == "insert") ced->insert_curve (); else {
			int i = atoi (cmd.c_str());
			if (i != 0) cl->move (i);
		}
	}	else {
		result = "no curve library";
		return false;
	}
	return true;
}

bool quit::operator() (tokenizer& tz) {
	extern int quit_now;
	quit_now = 1;
	return true;
}

extern uis_t uis;
bool set_curve_editor::operator() (tokenizer& tz) {
	string name;
	int screen;
	tz >> name >> screen;
	uis.set_editor (name, screen);
	return true;
}

bool get_curve_editor::operator() (tokenizer& tz) {
	string what;
	tz >> what;
	if (what == "n" || what == "num_screens") {
		sprintf (buf, "%d", uis_t::MAX_EDITORS);
		result = buf;
		return true;  
	} else {
		int i = atoi (what.c_str());
		if (i > 0) {
			ui* u = uis.ed[i-1];
			if (u) {
				result = u->name;
				return true;
			} else {
				result = "invalid_editor";
				return false;
			}
		}
		return false;
	}
	
}

bool midi::operator() (tokenizer& tz) {
	string cmd; tz >> cmd;
	if (cmd == "") {
		list_devices ();
		return true;
	} else if (cmd == "open" || cmd == "o") {
		int port;  tz >> port;
		int d = find_free_device ();
		if (d == -1) {
			result = "all devices in use. please close a device";
			return false;
		} else return open_port_on_device (d, port);
	}
	return false;
}

int midi::find_free_device () {
	for (int i = 0; i < MAX_DEVICES; ++i) {
		RtMidi& d = devices[i];
		if (!d.connected_) return i; 
	}
	return -1;
}

int midi::open_port_on_device (int dev, int port) {
	RtMidi& d = devices[dev];
	d.openPort (port);
	++nports;
	sprintf (buf, "opened port %d on device %d", port, dev);
	result = buf;
	return 1;
}

void midi::list_devices () {
	RtMidiIn& device = devices[0];
	int nports = device.getPortCount ();
	for (int i = 0; i < nports; ++i) {
		string name = device.getPortName (i);
		cons << console::cyan << "port_number " << i << " port_name " << name << eol;
	}
}

extern tcl_interp interpreter;

void midi::handle_input () {

	extern double MIDI_BPM;
	extern chrono clk;

	vector <unsigned char> msg;
	for (int i = 0, j = 0; (i < MAX_DEVICES) && (j < nports); ++i) {
		RtMidiIn& dev = devices[i];
		if (dev.connected_) {
			++j;
			dev.getMessage (&msg);
			int n = msg.size ();
			if (n) {
				unsigned char byte1 = msg[0];
				if (byte1 == 0xf8) { // midi clock signal
					static float start_time = clk.secs;
					static int num_clocks = 0;
					float elapsed_time = clk.secs - start_time;
					++num_clocks;
					if (elapsed_time >= 1) {
						float f_num_clocks = 1. / elapsed_time * num_clocks;
						double midi_bpm = f_num_clocks * 2.5;
						MIDI_BPM = midi_bpm;
						interpreter ("midi_clock");
						num_clocks = 0;
						start_time = clk.secs;
					}  
				} else if ((byte1 & 0xf0) == 0xb0) { // control change
					run ("midi_cc", msg);
				}			
			}
		}
	}
	
}

void midi::run (const string& c, const vector <unsigned char>& msg) {
	
	string cmd (c);
	for (int i = 0, j = msg.size(); i < j; ++i) {
			sprintf (buf, " %d", msg[i]);
			cmd += buf; 
	}
	interpreter (cmd);
	
}


