#include "solver.h"
#include "multi_curve.h"
#include <cmath>

solver::solver () {
  crv = 0;
  pts = 0;
  npts = 0;
  last = last_1 = 0;
  firstx = firsty = lastx = lasty = 0;
}

solver::solver (multi_curve* c) {
  operator() (c);
}

void solver::operator() (multi_curve* c) {
  crv = c;
	init ();
}

void solver::init () {
		
  pts = crv->get_points ();
  npts = crv->num_points ();

    
  // set current segment to first segment
	setseg (0, 1);
  
  // last, last but 1 points
  last = npts - 1;
  last_1 = last - 1;

  // find first and last x and its corresponding y
  // we assume x0 <= x1 <= x2 <= ..... <= xlast
  firstx = pts[0].x; firsty = pts[0].y;
  lastx = pts[last].x; lasty = pts[last].y;


}

float solver::operator() (float x) {
	if (x < startx || x > endx) {
		if (!findseg (x)) {
			if (x < firstx) return firsty; 
			else if (x > lastx) return lasty; 
			else return result;
		}
  }
	result = ycomp + m * x;
	return result;
}

// solves y for x with modulation of x
void solver::operator() (float& x, float& dx, int q, float* mod, float* y, xhandler& xmin, xhandler& xmax, int dp) {
	for (int p = 0; p < q; p += dp) {
		x += (mod[p] + dx);
		if (x < startx || x > endx) {
			if (!findseg (x)) {
				if (x > lastx) xmax (*this, x, dx); 
				else if (x < firstx) xmin (*this, x, dx);
			}
		} 
		y[p] = ycomp + m * x;
	}
}

// solves y for x
void solver::operator() (float& x, float& dx, int q, float* y, xhandler& xmin, xhandler& xmax, int dp) {
	for (int p = 0; p < q; p += dp) {
		x += dx;
		if (x < startx || x > endx) {
			if (!findseg (x)) {
				if (x > lastx) xmax (*this, x, dx); 
				else if (x < firstx) xmin (*this, x, dx);
			}
		} 
		y[p] = ycomp + m * x;
	}
}

// solves array of x. stores solution ie y in the same array at corresponding location
void solver::operator() (float* ax, int q, int dx, xhandler& xmin, xhandler& xmax) {
	
	float x;
	for (int p = 0; p < q; ++p, ax += dx) {
		x = *ax;
		if (x < startx || x > endx) {
			if (!findseg(x)) {
				if (x > lastx) {
					x -= lastx;
					setseg (0, 1);
					if (!inseg (x, 0, 1)) findseg (x);
				} else if (x < firstx) {
					x = firstx;
					setseg (0, 1);
				}
			}
		}
		*ax = ycomp + m * x;
	}

}

atmin _atmin;
tomin _tomin;
loopmin _loopmin;
loopmax _loopmax;
pongmin _pongmin;
pongmax _pongmax;

inline void atmin::operator() (solver& s, float& x, float& dx) {
	x = s.firstx;
	s.setseg (0, 1);
}

inline void tomin::operator() (solver& s, float& x, float& dx) {
	x -= s.lastx;
	s.setseg (0, 1);
	if (!s.inseg (x, 0, 1)) s.findseg (x);
}

inline void loopmin::operator() (solver& s, float& x, float& dx) {
	atmin::operator() (s, x, dx);
	if (dx < 0) dx = -dx;
}

inline void loopmax::operator() (solver& s, float& x, float& dx) {
	tomin::operator() (s, x, dx);
	if (dx < 0) dx = -dx;
}

inline void pongmax::operator () (solver& s, float& x, float& dx) {
	x = s.lastx;
	s.setseg (s.last_1, s.last);
	if (dx > 0) dx = -dx;
}

inline void pongmin::operator () (solver& s, float& x, float& dx) {
	x = s.firstx;
	s.setseg (0, 1);
	if (dx < 0) dx = -dx;
}
